# Original Copyright 2005 Dr Balwinder S Dheeman
# Modified and ported to rubygame by Immanuel Scholz
# Licensed under GPL

require "rubygame"
require "rubygame/tools"
include Rubygame

$screen = Screen.set_mode [270,270]
$backbuffer = Surface.new [270,270]
$clock.target_framerate = 30

class Board
  FILL = [64,64,64] # fill, gray
  NORM = [96,96,64] # normal, yellow
  OOPS = [128,0,0]  # invalid, red

  def initialize(size = 9)
    @size = size
    @box = Array.new(@size) { Array.new(@size, 0) }
    @level = -1;
    @x = 0; @y = 0
    @mask = [
#	[[],[],[],[],[]],
	[[0, 1, 2, 3, 5, 7, 8],[0, 1, 5, 6, 7],[0, 1, 3, 4, 7, 8],[1, 2, 3, 4, 6, 8],[0, 2, 3, 4, 5, 6, 8]],
	[[0, 3, 4, 5, 7, 8],[0, 1, 2, 3, 5, 6, 8],[0, 2, 4, 6, 8],[2, 3, 4, 5, 6, 7, 8],[1, 2, 6, 7]],
	[[0, 3, 4, 5, 7, 8],[0, 2, 5, 7],[0, 2, 4, 6, 8],[0, 1, 2, 3, 7],[1, 2, 4, 6, 7]],
	[[1, 2, 3, 4, 6, 7, 8],[0, 1, 2, 3, 5, 7, 8],[0, 2, 4, 6, 8],[0, 3, 4, 5, 6],[0, 1, 4, 7, 8]],
	[[1, 2, 3, 7, 8],[1, 2, 4, 5, 6, 8],[0, 1, 4, 5, 6, 7, 8],[0, 3, 4, 5, 6, 8],[1, 2, 3, 5, 6, 7]],
	[[0, 2, 3, 4, 5, 6, 7],[0, 1, 3, 4, 5, 8],[1, 4, 5, 7],[0, 1, 3, 4, 6],[0, 1, 2, 3, 5, 6, 7, 8]],
	[[0, 2, 4, 5, 6],[0, 1, 2, 3, 5, 8],[0, 3, 6, 7, 8],[1, 2, 3, 4, 5, 7, 8],[0, 4, 8]],
	[[0, 1, 2, 3, 4, 7],[0, 3, 4, 6, 8],[0, 1, 3, 4, 6, 8],[1, 2, 3, 6, 7],[1, 3, 4, 5, 7]],
	[[0, 1, 2, 4, 7, 8],[0, 2, 3, 4, 5, 6, 8],[1, 3, 4, 5, 8],[0, 1, 4, 6, 8],[0, 1, 3, 4, 5, 7, 8]],
	[[0, 1, 2, 3, 5, 6, 8],[1, 2, 3, 4, 6, 7],[2, 4, 5, 7, 8],[0, 1, 4, 5, 6, 7],[0, 3, 4, 5, 8]],
	[[0, 1, 2, 3, 5, 8],[0, 1, 3, 4, 5, 6, 7, 8],[0, 1, 2, 5, 7],[2, 4, 7, 8],[0, 2, 3, 4, 5, 6, 8]],
	[[1, 3, 4, 7, 8],[0, 2, 4, 5, 7, 8],[0, 2, 4, 5, 7, 8],[1, 2, 5, 6, 7],[2, 3, 4, 5, 6]],
	[[0, 1, 3, 4, 6, 8],[2, 3, 4, 6, 8],[0, 1, 2, 5, 6, 7],[2, 3, 4, 5, 7, 8],[0, 1, 3, 5, 7, 8]],
	[[2, 3, 4, 5, 6, 8],[3, 4, 5, 6],[0, 1, 2, 6, 7],[0, 1, 3, 4, 7, 8],[0, 1, 7, 8]],
	[[0, 1, 3, 4, 5, 6],[0, 1, 2, 3, 4, 6, 7, 8],[0, 1, 2, 5, 6, 8],[0, 2, 3, 5, 6, 7],[1, 2, 4, 6, 7]],
	[[0, 1, 2, 3, 4, 5, 6, 7, 8],[1, 2, 6, 7],[1, 4, 7],[0, 1, 4, 7, 8],[0, 2, 3, 4, 5, 6, 8]],
	[[0, 2, 3, 4, 5, 6, 7, 8],[0, 1, 2, 5, 6, 8],[0, 1, 2, 4, 5, 6, 7, 8],[1, 3, 4, 5, 7, 8],[0, 1, 3, 5, 7, 8]],
	[[1, 2, 4, 6],[0, 2, 4, 6, 7, 8],[0, 1, 3, 4, 5, 6, 7, 8],[0, 4, 5, 6, 7, 8],[1, 4, 7]],
	[[0, 1, 3, 4, 5],[0, 3, 4, 6, 8],[0, 2, 4, 5, 6, 7],[0, 1, 3, 4, 6, 7, 8],[1, 2, 4, 6, 7]],
	[[1, 4, 5, 6, 7, 8],[1, 2, 4, 5, 6],[0, 2, 4, 5, 6, 7, 8],[0, 1, 3, 4, 7],[2, 3, 4, 5, 6]],
	[[1, 2, 3, 4],[2, 4, 5, 6, 8],[0, 1, 2, 3, 5, 7],[1, 2, 3, 4, 5, 6, 8],[1, 2, 3, 4, 5, 6, 7]],
	[[1, 2, 4, 5, 7],[0, 1, 5, 6],[0, 4, 5, 6, 7, 8],[0, 4, 5, 6, 7],[0, 2, 4, 6, 8]],
	[[0, 3, 5, 7],[0, 2, 3, 6, 7, 8],[0, 1, 4, 5, 8],[2, 3, 4, 6, 7],[0, 1, 4, 7, 8]],
	[[0, 1, 2, 5, 7],[1, 2, 3, 4, 6, 8],[0, 2, 3, 4, 5, 6, 7, 8],[3, 4, 5, 6, 7],[0, 1, 3, 4, 5, 7, 8]],
	[[0, 1, 2, 4, 5, 6, 7, 8],[3, 4, 5, 7, 8],[1, 2, 4, 5, 8],[0, 1, 2, 4, 5, 6, 8],[0, 2, 6, 8]],
	[[3, 4, 5, 6, 7, 8],[2, 4, 8],[0, 1, 2, 3, 4, 6, 7],[0, 1, 3, 4, 5, 6, 7, 8],[0, 2, 4, 6, 8]],
	[[0, 3, 4, 5, 6, 7, 8],[1, 2, 6, 7, 8],[0, 1, 2, 3, 5],[0, 1, 3, 4, 5],[0, 1, 2, 4, 6, 7, 8]],
	[[1, 2, 3, 5, 6, 7],[0, 4, 5, 6, 7],[0, 1, 2, 6, 7],[3, 5, 6, 7, 8],[0, 2, 3, 5, 6, 8]],
	[[3, 7, 8],[0, 1, 2, 4, 5, 6, 7, 8],[0, 1, 2, 7, 8],[3, 4, 6, 7, 8],[2, 3, 4, 5, 6]],
	[[0, 2, 4, 6, 8],[0, 1, 4, 7, 8],[1, 2, 3, 4, 5, 6, 7],[1, 2, 4, 6, 7],[0, 1, 3, 4, 5, 7, 8]]]
  end

  def newgame
    shuffle
    mask
    @game_over_shown = false
  end

  def redraw
    $backbuffer.fill [0,0,0]
    @remain = 0
    for y1 in 0 .. @size - 1
      for x1 in 0 .. @size - 1
        bg = bkground(x1, y1)
        $backbuffer.fill bg, [x1*30, y1*30, 30, 30]
        if @box[y1][x1] == 0
          @remain += 1
        else
          n, col = @box[y1][x1], [255,255,255]
          if n > @size
            n -= @size
            col = [255,128,128]
          end
          $backbuffer.draw_string n, :x => x1*30+11, :y => y1*30+5, :color => col
        end
      end
    end
    $backbuffer.draw_box [@x*30, @y*30], [@x*30+29, @y*30+29], [255,0,0]
  end
  
  def left
    @x -= 1 if @x > 0
  end

  def right
    @x += 1 if @x < @size - 1
  end

  def up
    @y -= 1 if @y > 0
  end

  def down
    @y += 1 if @y < @size - 1
  end

  def legal? num
    for x1 in 0...@size
      return false if @box[@y][x1].abs % (@size + 1) == num
      return false if @box[x1][@x].abs % (@size + 1) == num
    end
    xo,yo = @x/3*3, @y/3*3 # quadrant
    for y in 0..2
      for x in 0..2
        return false if @box[y+yo][x+xo] == num
      end
    end
    true
  end

  def fill(num)
    num += @size unless legal? num
    if @box[@y][@x] < 1 || @box[@y][@x] > @size
      if num >= 1 && num <= @size
        @remain -= 1
      end
      @box[@y][@x] = num
    end
  end

  def clear
    @box[@y][@x] = @orig[@y][@x]
  end
  
  def mask
    size = @size - 1
    for x1 in 0 .. @size / 2
      for y1 in 0 .. @size - 1
        if @mask[@level][x1].index(y1)
          @box[x1][y1] = 0
          @box[size - x1][size - y1] = 0
        end
      end
    end
    @orig = @box.map {|line| line.dup} # deep copy
  end

  def over
    return if @game_over_shown
    grayout
    $backbuffer.draw_string "*** YOU WIN !! ***", :x => 85, :y => 130
    $pause = true
    @game_over_shown = true
  end

  def over?
    return @remain == 0 ? true : false
  end

  private
  def bkground(x1, y1)
    if [0, 1, 2, 6, 7, 8].index(y1) and [0, 1, 2, 6, 7, 8].index(x1)
      FILL
    elsif [3, 4, 5].index(y1) and [3, 4, 5].index(x1)
      FILL
    else
      NORM
    end
  end

  def shuffle
    num = rand( @size)
    for x1 in 0 .. @size - 1
      num += 1 if x1 % (@size / 3) == 0
      num -= @size if num > @size
      for y1 in 0 .. @size - 1
        @box[x1][y1] = num
        num += 1
        num -= @size if num > @size
      end
      num += 3
    end

    n1 = rand((@size + 1) / 3); n2 = (n1 + 1) % 3
    size = @size - 1
    @box[n1], @box[n2] = @box[n2], @box[n1]
    @box[n1 + 3], @box[n2 + 3] = @box[n2 + 3], @box[n1 + 3]
    @box[size - n1], @box[size - n2] = @box[size - n2], @box[size - n1]
    n1 = rand((@size + 1) / 3); n2 = (n1 + 1) % 3
    for x1 in 0 .. @size - 1
      @box[x1][n1], @box[x1][n2] = @box[x1][n2], @box[x1][n1]
      @box[x1][n1 + 3], @box[x1][n2 + 3] = @box[x1][n2 + 3], @box[x1][n1 + 3]
      @box[x1][size - n1], @box[x1][size - n2] = @box[x1][size - n2], @box[x1][size - n1]
    end
    @level = rand( @mask.size)
    @level = (@level + 1) % @mask.size if $DEBUG
  end
end


def paused
  old = $pause
  $pause = false
  old
end

def grayout
  gray = Surface.new [270,270]
  gray.set_alpha 128
  gray.blit $backbuffer, [0,0]
end

def help
  grayout
  txt = <<-EOB
    Fill the board. Every number only 
    once in each row, column and 3x3 block.
  
    F1|This help
    cursor keys|Move cursor around
    space|Undo the current field
    1 to 9|Place the number
    r|Restart with current Sudoku
    n|Restart with a new Sudoku
    Escape|Quit
  EOB
  y = 40
  txt.each do |line|
    t1,t2 = line.chomp.strip.split "|"
    $backbuffer.draw_string t1 || " ", :x => 20, :y => y
    $backbuffer.draw_string t2 || " ", :x => 90, :y => y
    y += 20
  end
  $pause = !$pause
end

bd = Board.new
bd.newgame

register :key => K_N do paused or bd.newgame end
register :key => K_R do paused or bd.mask end
register :key => K_DOWN do paused or bd.down end
register :key => K_UP do paused or bd.up end
register :key => K_LEFT do paused or bd.left end
register :key => K_RIGHT do paused or bd.right end
register :key => K_SPACE do paused or bd.clear end
register :key => /^[1-9]$/ do |e| paused or bd.fill Key::KEY2ASCII[e.key].to_i end
register :key => K_ESCAPE do paused or throw :quit end
register :key => K_F1 do paused or help end

register_on_tick do
  bd.redraw unless $pause
  bd.over if bd.over?
end

start
